/*
  ==============================================================================

    This file contains the basic framework code for a JUCE plugin processor.

  ==============================================================================
*/

#include "PluginProcessor.h"
#include "PluginEditor.h"

//==============================================================================
FinalSamplerAudioProcessor::FinalSamplerAudioProcessor()
#ifndef JucePlugin_PreferredChannelConfigurations
    : AudioProcessor(BusesProperties()
#if ! JucePlugin_IsMidiEffect
#if ! JucePlugin_IsSynth
        .withInput("Input", juce::AudioChannelSet::stereo(), true)
#endif
        .withOutput("Output", juce::AudioChannelSet::stereo(), true)
#endif
    ), myTreeState(*this, nullptr, "Parameters", createParameterLayout())//construct tree state (thisprocessor, noundomanager, name of paramlayout, create)
#endif
{
    //call registerbasicformats. Juce includes wav and mp3. Necessary to actually stream audio
    myFormatManager.registerBasicFormats();

    //set up listener in (this) processor. using treevalue so I dont have to have a listener for each param
    myTreeState.state.addListener(this);


    //add voices to sampler
    for (int i = 0; i < numberOfSamplerVoices; i++) {
        mySampler.addVoice(new juce::SamplerVoice());
    }
}

FinalSamplerAudioProcessor::~FinalSamplerAudioProcessor()
{
    //reset format reader and numsamplersounds
    myFormatReader = nullptr;

}

//==============================================================================
const juce::String FinalSamplerAudioProcessor::getName() const
{
    return JucePlugin_Name;
}

bool FinalSamplerAudioProcessor::acceptsMidi() const
{
#if JucePlugin_WantsMidiInput
    return true;
#else
    return false;
#endif
}

bool FinalSamplerAudioProcessor::producesMidi() const
{
#if JucePlugin_ProducesMidiOutput
    return true;
#else
    return false;
#endif
}

bool FinalSamplerAudioProcessor::isMidiEffect() const
{
#if JucePlugin_IsMidiEffect
    return true;
#else
    return false;
#endif
}

double FinalSamplerAudioProcessor::getTailLengthSeconds() const
{
    return 0.0;
}

int FinalSamplerAudioProcessor::getNumPrograms()
{
    return 1;   // some hosts don't cope very well if you tell them there are 0 programs,
                // so this should be at least 1, even if you're not really implementing programs.
}

int FinalSamplerAudioProcessor::getCurrentProgram()
{
    return 0;
}

void FinalSamplerAudioProcessor::setCurrentProgram(int index)
{
}

const juce::String FinalSamplerAudioProcessor::getProgramName(int index)
{
    return {};
}

void FinalSamplerAudioProcessor::changeProgramName(int index, const juce::String& newName)
{
}

//==============================================================================
void FinalSamplerAudioProcessor::prepareToPlay(double sampleRate, int samplesPerBlock)
{
    // Use this method as the place to do any pre-playback
    // initialisation that you need..

    //SET SAMPLE RATE ON SAMPLER OBJECT
    mySampler.setCurrentPlaybackSampleRate(sampleRate);

    //update parameters before anything
    updateADSR();


}

void FinalSamplerAudioProcessor::releaseResources()
{
    // When playback stops, you can use this as an opportunity to free up any
    // spare memory, etc.
}

#ifndef JucePlugin_PreferredChannelConfigurations
bool FinalSamplerAudioProcessor::isBusesLayoutSupported(const BusesLayout& layouts) const
{
#if JucePlugin_IsMidiEffect
    juce::ignoreUnused(layouts);
    return true;
#else
    // This is the place where you check if the layout is supported.
    // In this template code we only support mono or stereo.
    // Some plugin hosts, such as certain GarageBand versions, will only
    // load plugins that support stereo bus layouts.
    if (layouts.getMainOutputChannelSet() != juce::AudioChannelSet::mono()
        && layouts.getMainOutputChannelSet() != juce::AudioChannelSet::stereo())
        return false;

    // This checks if the input layout matches the output layout
#if ! JucePlugin_IsSynth
    if (layouts.getMainOutputChannelSet() != layouts.getMainInputChannelSet())
        return false;
#endif

    return true;
#endif
}
#endif

void FinalSamplerAudioProcessor::processBlock(juce::AudioBuffer<float>& buffer, juce::MidiBuffer& midiMessages)
{
    //below is channel setup provided by juce
    juce::ScopedNoDenormals noDenormals;
    auto totalNumInputChannels = getTotalNumInputChannels();
    auto totalNumOutputChannels = getTotalNumOutputChannels();

    //clear out garbage data
    for (auto i = totalNumInputChannels; i < totalNumOutputChannels; ++i)
        buffer.clear(i, 0, buffer.getNumSamples());

    //get values of parameters from valuetreestate
    auto& attack = *myTreeState.getRawParameterValue("ATTACK");
    auto& decay = *myTreeState.getRawParameterValue("DECAY");
    auto& sustain = *myTreeState.getRawParameterValue("SUSTAIN");
    auto& release = *myTreeState.getRawParameterValue("RELEASE");
    //return std::atomic<float>*, we need const floats so we dereference

    //if a parameter has been changed, update. (maybe try to do without bool)
    if (shouldBeUpdated == true) {
        updateADSR();
    }

    //deprecated can not use
           // juce::MidiBuffer::Iterator myMidiBufferIterator {midiMessages};


    //need iterator to pull messages out of midibuffer and set boolean accordingly
   //new midibufferIterator class allows easier iterating using range-for
    for (const juce::MidiMessageMetadata metadata : midiMessages) {
        // declare midimessage to store from midibuffer. for playhead, need to know where in buffer I'm at to track along
        const juce::MidiMessage pulledMidiMessage = metadata.getMessage();
        if (pulledMidiMessage.isNoteOn()) {
            noteIsBeingPlayed = true;
            //to handle note change in same block
           // sampleCount = 0;
            //find a way to get only most recently pressed on
        }
        else if (pulledMidiMessage.isNoteOff()) {
            noteIsBeingPlayed = false;
            //to handle note change in same block
            sampleCount = 0;
        }
    }

    //want to know how long note has been played for. counted in samples. processblock deals with buffer.getnumsamples amount of info each time 
    //when note is being played, start a counter. stop it on note off
        //if note is being played, increment samplecount by buffersize. once note is off, return to 0


    sampleCount = noteIsBeingPlayed ? sampleCount += buffer.getNumSamples() : 0;    // get index of current sample dont use whole block
     //if (noteIsBeingPlayed == true) sampleCount += buffer.getNumSamples();

     //call rendernextblock that parses through midimessages for us
    mySampler.renderNextBlock(buffer, midiMessages, 0, buffer.getNumSamples());
}

//==============================================================================
bool FinalSamplerAudioProcessor::hasEditor() const
{
    return true; // (change this to false if you choose to not supply an editor)
}

juce::AudioProcessorEditor* FinalSamplerAudioProcessor::createEditor()
{
    return new FinalSamplerAudioProcessorEditor(*this);

    //when FinalSamplerAudioProcessor gets instantiated, processor gives a reference to itself as argument

}

//==============================================================================
void FinalSamplerAudioProcessor::getStateInformation(juce::MemoryBlock& destData)
{
    // You should use this method to store your parameters in the memory block.
    // You could do that either as raw data, or use the XML or ValueTree classes
    // as intermediaries to make it easy to save and load complex data.
}

void FinalSamplerAudioProcessor::setStateInformation(const void* data, int sizeInBytes)
{
    // You should use this method to restore your parameters from this memory block,
    // whose contents will have been created by the getStateInformation() call.
}

//==============================================================================
// This creates new instances of the plugin..
juce::AudioProcessor* JUCE_CALLTYPE createPluginFilter()
{
    return new FinalSamplerAudioProcessor();


}

void FinalSamplerAudioProcessor::loadFile() {
    //clear sounds or it will just keep adding
    mySampler.clearSounds();

    //use file chooser clase and construct it with dialog box title. Actually opens the browser dialog
    juce::FileChooser myFileChooser{ "Please choose a file to load" };

    // basically "if file has been chosen"
    if (myFileChooser.browseForFileToOpen() == true) {
        //retrieve user's selection
        auto fileToLoad = myFileChooser.getResult();

        //create reader for selected file
        myFormatReader = myFormatManager.createReaderFor(fileToLoad);
    }

    //need BigInteger to represent range of midi notes I want sound to be able to be played with (i want all)
    juce::BigInteger midiRange;
    //set range for biginteger bits to be set to true, meaning I want all 128 midi notes to be able to play the sampled audio
    midiRange.setRange(0, 128, true);


    //Add a new sound to the sampler.
    mySampler.addSound(new juce::SamplerSound("Sample", *myFormatReader, midiRange, 60, 0.1, 0.1, 12.0));
    //The object passed into addSound() is reference counted, so it will be deleted when the synthesiser and all voices are no longer using it.

       //SamplerSound(const String & name, AudioFormatReader & source, const BigInteger & midiNotes, 
       // int midiNoteForNormalPitch, double attackTimeSecs, double releaseTimeSecs, double maxSampleLengthSeconds);
                //need to dereference formatreader   
                //midinotefornormalpitch. Typically is C3, which is midinote 60
        //Creates a sampled sound from an audio reader. This will attempt to load the audio from the source 
        //into memoryand store it in this object.

        //make sure adsr applied to loaded file 
    updateADSR();

}

//drag and drop load file
void FinalSamplerAudioProcessor::loadFile(const juce::String& filePath) {
    //clear sounds or it will just keep adding
    mySampler.clearSounds();

    //path is given to processor by eidtor via drag and drop classs
    //create juce file object from path
    juce::File fileToLoad = juce::File(filePath);

    //create reader for selected file
    myFormatReader = myFormatManager.createReaderFor(fileToLoad);

    //MAKE SURE WAVEFORM BUFFER HAS SAME SIZE AS READER
    auto numberOfSamples = static_cast<int>(myFormatReader->lengthInSamples);
    waveformBuffer.setSize(1, numberOfSamples);

    //read sample into buffer (using only left channel read for simplicity)
    myFormatReader->read(&waveformBuffer, 0, numberOfSamples, 0, true, false);


    //need BigInteger to represent range of midi notes I want sound to be able to be played with (i want all)
    juce::BigInteger midiRange;
    //set range for biginteger bits to be set to true/1, meaning we want all 128 midi notes to be able to play the sampled audio
    midiRange.setRange(0, 128, true);


    //Add a new sound to the sampler.
    mySampler.addSound(new juce::SamplerSound("Sample", *myFormatReader, midiRange, 60, 0.1, 0.1, 12.0));

}

/*formatReader->read can take in an audio buffer. I tell it where in readersampleand audio buffer i want it to start from.
    it puts the sample info into audio buffer

 read(AudioBuffer< float >* buffer, int startSampleInDestBuffer, int numSamples, int64 readerStartSample,
        bool useReaderLeftChan, bool useReaderRightChan)
        *int64 = long long (juce math, platform-independent 64-bit integer type)

        Read converts the reader's fixed- or floating-point data to the buffer's floating-point format, and will
        try to deal with mismatches between the number of channels in the reader and the buffer.

 */

void FinalSamplerAudioProcessor::updateADSR() {

    //set parameters with values from tree state, which were loaded by the slider
    myAdsrParameters.attack = myTreeState.getRawParameterValue("ATTACK")->load();
    myAdsrParameters.decay = myTreeState.getRawParameterValue("DECAY")->load();
    myAdsrParameters.sustain = myTreeState.getRawParameterValue("SUSTAIN")->load();
    myAdsrParameters.release = myTreeState.getRawParameterValue("RELEASE")->load();
    //load bc atomic float pointers. arrow bc its a pointer, load to give me value

    //iterate through sounds in sampler
    for (int soundIndex = 0; soundIndex < mySampler.getNumSounds(); soundIndex++) {

        mySampler.getSound(soundIndex);
        //sampler is technically a synth object. we need to only deal with sampler sounds, even though synths have both
        // only sample sounds have updateEnvelope. without dynamic cast, it cant know for sure
            //dynamic cast as samplersound*. mysampler.getsound().get to get pointer
        if (auto thisSound = dynamic_cast<juce::SamplerSound*>(mySampler.getSound(soundIndex).get())) {
            
            //apply values to sampler sound with native juce method
            thisSound->setEnvelopeParameters(myAdsrParameters);
        }
    }


}



juce::AudioProcessorValueTreeState::ParameterLayout FinalSamplerAudioProcessor::createParameterLayout() {

    //make vector of parameters (ranged audio parameters)
    std::vector<std::unique_ptr<juce::RangedAudioParameter>> parameters;

    // add ADSR stuff
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("ATTACK", "Attack", juce::NormalisableRange<float> {0.1f, 1.0f, 0.01f }, 0.1f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("DECAY", "Decay", juce::NormalisableRange<float> {0.1f, 2.0f, 0.01f }, 0.1f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("SUSTAIN", "Sustain", juce::NormalisableRange<float> {0.1f, 1.0f, 0.01f }, 1.0f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("RELEASE", "Release", juce::NormalisableRange<float> {0.1f, 3.0f, 0.01}, 0.6f));


    return{ parameters.begin(), parameters.end() };
}

//virtual listener funct
void FinalSamplerAudioProcessor::valueTreePropertyChanged(juce::ValueTree& treeWhosePropertyHasChanged, const juce::Identifier& propertyThatHasChanged) {
    //this gets called with processor constructor, which constructs tree state, which needs a parameter layout
    shouldBeUpdated = true;

}