/*
  ==============================================================================

    WaveformThumbnail.cpp
    Created: 1 Aug 2021 5:37:04pm
    Author:  nweihsma

  ==============================================================================
*/

#include <JuceHeader.h>
#include "WaveformThumbnailComponent.h"

//==============================================================================
WaveformThumbnailComponent::WaveformThumbnailComponent(FinalSamplerAudioProcessor& p) : myAudioProcessor(p)
//tell FinalSamplerAudioProcessor& member that the processor I want to use the argument from main constructor
{
    // In your constructor, you should add any child components, and
    // initialise any special settings that your component needs.

}

WaveformThumbnailComponent::~WaveformThumbnailComponent()
{
}

void WaveformThumbnailComponent::paint(juce::Graphics& g)
{
    //set thumbnail to nmu colors to differentiate from background
    g.fillAll(juce::Colours::darkgreen.darker(0.6f));
    g.setColour(juce::Colours::gold);

    // if audio file has been loaded.  as long as we have a sampler sound loaded
    if (myAudioProcessor.getNumberOfSamplerSounds() > 0) {
        //create an empty path for drawing waveform easily
        juce::Path waveformPath;


        //clear out any previous points
        pointsToDrawWaveform.clear();

        //if file has been dropped onto plugin, get waveform from processor
        auto waveFormBuffer = myAudioProcessor.getWaveformBuffer();
        //get read pointer
        auto bufferRead = waveFormBuffer.getReadPointer(0);

        //scale for length of audio. 1 sec = 44100 samples, plugin width only so many pixels 
        //to draw evenly spaced samples, getting as close as possible to actual waveform
        auto xScaledRatio = waveFormBuffer.getNumSamples() / getWidth();

        //iterate through waveform and scale X axis
        for (int sampleIndex = 0; sampleIndex < waveFormBuffer.getNumSamples(); sampleIndex += xScaledRatio) {
            //add sample to scaled buffer to draw
            pointsToDrawWaveform.push_back(bufferRead[sampleIndex]);

        }
        //basically starts a line (putting pencil down at a point) 
        //0 to be at left of graph, half the height because the waveform will start from 0 (middle of graph)
        waveformPath.startNewSubPath(0, getHeight() / 2);

        //scale Y axis of waveform. I have (waveformbuffer / getwidth()) values between -1 and +1. Need to scale up
        for (int sampleIndex = 0; sampleIndex < pointsToDrawWaveform.size(); sampleIndex++) {

            // pointsToDrawWaveform is a std vector, not audio buffer so I dont need a write pointer
                 //mapping the highest value in scale to 0, to correspond with y coordinate 0 in juce component (top of screen)

            //jmap(sourceValue, sourceRangeMin, sourceRangeMax, targetRangeMin, targetRangeMax)
            //template argument is float
            auto waveformDrawPoint = juce::jmap<float>(pointsToDrawWaveform[sampleIndex], -1.0f, 1.0f, getHeight(), 0); 
            //jmap returns the type i give it (float)

            //draw the line after pen is put down. sample index = x value, drawpoint = flipped & scaled y value
            waveformPath.lineTo(sampleIndex, waveformDrawPoint);
        }
        //before filling, line has not been technically drawn, only mapped out.
            //stroke type is very basic as of now, only given a width
        g.strokePath(waveformPath, juce::PathStrokeType(3));

        //draw the form with a juce path   
        //juce path is sequence of lines and curves that can make a closed shape or be open-ended  
        //Create empty path, add lines and curves to it to create shapes, 
        //then it can be rendered by my Graphics context.
        //Useful because it can take group of points and draw at once. 

        //write file name
        g.setColour(juce::Colours::antiquewhite);
        g.setFont(16.0f);

        //set a bounds for file name
        auto fileNameBounds = getLocalBounds().reduced(8);

        g.drawFittedText(myFileName, fileNameBounds, juce::Justification::bottomRight, 1);
        //	drawFittedText (const String &text, Rectangle< int > area, Justification justificationFlags, int maximumNumberOfLines, float minimumHorizontalScale=0.0f)

        //map sample number to width of screen
         //jmap(sourceValue, sourceRangeMin, sourceRangeMax, targetRangeMin, targetRangeMax)
       // min = beginning of sample, max = numsamples in audio file
            //template argument of type int
        //sample count will get mapped to number between 0 and get width. this is the int return
        auto playheadDrawPoint = juce::jmap<int>(myAudioProcessor.getSampleCount(), 0, myAudioProcessor.getWaveformBuffer().getNumSamples(), 0, getWidth());//return int

        //draw playhead line (startx, starty, endx, endy, line thickness)
            //vertical line at playhead, full height of component, thin but thick enough to easily see
        g.drawLine(playheadDrawPoint, 0, playheadDrawPoint, getHeight(), 3.0f);

        //draw transparent black rentangle behind playhead to show movement
        g.setColour(juce::Colours::black.withAlpha(0.25f));
        g.fillRect(0, 0, playheadDrawPoint, getHeight());

    }
    else {
        g.setFont(20.0f);
        g.drawFittedText("Drag and Drop an Audio File to Load", getLocalBounds(), juce::Justification::centred, 1);
    }

}

void WaveformThumbnailComponent::resized()
{
    // This method is where you should set the bounds of any child
    // components that your component contains..

}



//this automatically calls when detects that user has tried to drag something to it with mouse
//gives string array of paths of files that user has dragged onto it 
bool WaveformThumbnailComponent::isInterestedInFileDrag(const juce::StringArray& fileArray) {
    //check if files are valid (audio files)
        //iterate through filearray
    for (auto file : fileArray) {
        //look at filenames for extensions
        if (file.contains(".wav") || file.contains(".mp3") || file.contains(".aif")) {
            //user is attempting to drag valid file, I am interested in the file drag
            return true;
        }
    }

    //no files are valid, i am not interested
    return false;

}

void WaveformThumbnailComponent::filesDropped(const juce::StringArray& fileArray, int x, int y) {
    //iterate through file array to check if interested
    for (auto file : fileArray) {
        if (isInterestedInFileDrag(fileArray) == true) {
            //use juce file object to get filename for display instead of regex
                //make_unique pointer to juce file obj, not unque_ptr because I am allocating too 
            auto myLoadedFile = std::make_unique<juce::File>(file);
            //set string member to value from pointer method
            myFileName = myLoadedFile->getFileNameWithoutExtension();

            //paint the wave form
            shouldPaintWaveform = true;
            //load file from path using processor method
            myAudioProcessor.loadFile(file);
        }

    }

    repaint();
}

