
#include <JuceHeader.h>
#include "FilterComponent.h"

//==============================================================================
FilterComponent::FilterComponent(juce::String componentName, juce::AudioProcessorValueTreeState& treeState, juce::String selectorID, juce::String filterResonanceParameterID,
    juce::String filterCutoffParameterID)
{
    myComponentName = componentName;

    //set up juce string array for state variable filter type Choices
    juce::StringArray filterChoices{ "Lowpass", "Bandpass", "Highpass" };
    filterTypeSelector.addItemList(filterChoices, 1);//1 is ID offset

    addAndMakeVisible(filterTypeSelector);



    filterTypeSelectorAttachment = std::make_unique<juce::AudioProcessorValueTreeState::ComboBoxAttachment>(treeState, selectorID, filterTypeSelector);
    //user choice connecting to visual representation
    filterSelectorLabel.attachToComponent(&filterTypeSelector, true);
    addAndMakeVisible(filterSelectorLabel);


    //set up sliders and labels
    sliderAndLabelSetup(treeState, filterCutoffSliderAttachment, filterCutoffFrequencySlider, filterCutoffSliderLabel, filterCutoffParameterID);
    sliderAndLabelSetup(treeState, filterResonanceSliderAttachment, filterResonanceSlider, filterResonanceSliderLabel, filterResonanceParameterID);

    
    //user makes choice, signals processor audiovaluetreestate that change has been made. In audio callback (processor process block)
    //voice->getosc() is called in process block and sets new wave type
    filterCutoffFrequencySlider.setTextValueSuffix(" Hz");

}

FilterComponent::~FilterComponent()
{
}

void FilterComponent::paint(juce::Graphics& g)
{
    auto bounds = getLocalBounds().reduced(3);
    auto filterLabelSpace = bounds.removeFromTop(16.0f);


    g.setColour(juce::Colours::white);
    g.setFont(18.0f);
    g.drawText(myComponentName, filterLabelSpace, juce::Justification::centred);

    g.drawRoundedRectangle(bounds.getX(), bounds.getY() + 6.0f, bounds.getWidth(), bounds.getHeight() - 8.0f, 4.0f, 2.0f);

    g.setColour(juce::Colours::white);
    g.setFont(15.0f);
}

void FilterComponent::resized()
{
    //var to hold bounds of whole plugin, minus a bit
    const auto bounds = getLocalBounds().reduced(10);
    //var for distance between sliders
    const float sliderPadding = 10;
    //var for spacing slider evenly in box
    const float sliderWidth = (bounds.getWidth() / 2) - sliderPadding;
    const float sliderHeight = (bounds.getHeight() * 0.65); 

    //x,y point for first slider
    const float sliderStartPointX = bounds.getX();
    const float sliderStartPointY = (filterTypeSelector.getBottom() + 20.0f);


    filterTypeSelector.setBounds(bounds.getX() + bounds.getWidth() / 4, 30, bounds.getWidth() / 2, 20);

    //set bounds of sliders
    filterResonanceSlider.setBounds(sliderStartPointX, sliderStartPointY, sliderWidth, sliderHeight);
    //set other sliders using previous's x position 
    filterCutoffFrequencySlider.setBounds(filterResonanceSlider.getRight() + 10, sliderStartPointY, sliderWidth, sliderHeight);

    //draw labels relative to sliders (center bottom)
    filterResonanceSliderLabel.setBounds(filterResonanceSlider.getX(), filterResonanceSlider.getY() - 15, filterResonanceSlider.getWidth(), 15);
    filterCutoffSliderLabel.setBounds(filterCutoffFrequencySlider.getX(), filterCutoffFrequencySlider.getY() - 15, filterCutoffFrequencySlider.getWidth(), 15);

}
void FilterComponent::sliderAndLabelSetup(juce::AudioProcessorValueTreeState& treeState, std::unique_ptr<juce::AudioProcessorValueTreeState::SliderAttachment>& sliderAttachment,
    juce::Slider& slider, juce::Label& label, const juce::String parameterID) {
    //slider attachments

    //initialize with treestate etc
    sliderAttachment = std::make_unique<juce::AudioProcessorValueTreeState::SliderAttachment>(treeState, parameterID, slider);
    //===================================================================================================
    //slider

    //style sliders
    slider.setSliderStyle(juce::Slider::SliderStyle::RotaryHorizontalVerticalDrag);

    //set rotary parameters to change draw angle 
    ////rotary params radians measured from top of circle, instead of right like usual
    //// need larger end angle or else slider drawn in reverse, 
    float newStartAngleRadians = juce::MathConstants<float>::pi * 0.0f;
    float newEndAngleRadians = juce::MathConstants<float>::pi * 1.5f;

    slider.setRotaryParameters(newStartAngleRadians, newEndAngleRadians, true);




    //set text boxes
    slider.setTextBoxStyle(juce::Slider::TextBoxBelow, true, 60, 25);


    //add and make visible
    addAndMakeVisible(slider);
    //===================================================
    //label

    //set text
    label.setFont(17.0f);
    //attach to slider
    label.attachToComponent(&slider, true);
    //set justification type to make sure it is in middle
    label.setJustificationType(juce::Justification::centred);
    //make visible
    addAndMakeVisible(label);

}