

#include <JuceHeader.h>
#include "OscillatorComponent.h"

//==============================================================================
OscillatorComponent::OscillatorComponent(juce::AudioProcessorValueTreeState& treeState, juce::String selectorID, juce::String fmDepthParameterID,
    juce::String fmFrequencyParameterID)
{
    //set up wave choice array
    juce::StringArray waveChoices{ "Sine", "Square", "Saw" };

    //initalialise primary oscillator selector
    mainOscillatorWaveSelector.addItemList(waveChoices, 1);//1 is ID offset
    addAndMakeVisible(mainOscillatorWaveSelector);
    //attachments necessary for user choice connecting to visual representation
    mainWaveSelectorAttachment = std::make_unique<juce::AudioProcessorValueTreeState::ComboBoxAttachment>(treeState, selectorID, mainOscillatorWaveSelector);
    mainWaveSelectorLabel.attachToComponent(&mainOscillatorWaveSelector, true);
    addAndMakeVisible(mainWaveSelectorLabel);

    //user makes choice, signals processor audiovaluetreestate that change has been made. In audio callback (processor process block)
    //voice->getosc() is called in process block and sets new wave type

    //initalialise FM oscillator selector
    fmOscillatorWaveSelector.addItemList(waveChoices, 1);//1 is ID offset
    addAndMakeVisible(fmOscillatorWaveSelector);
    fmWaveSelectorAttachment = std::make_unique<juce::AudioProcessorValueTreeState::ComboBoxAttachment>(treeState, "FMWAVETYPE", fmOscillatorWaveSelector);
    fmWaveSelectorLabel.attachToComponent(&fmOscillatorWaveSelector, true);
    addAndMakeVisible(fmWaveSelectorLabel);



    //set up sliders and labels
    sliderAndLabelSetup(treeState, fmDepthSliderAttachment, fmDepthSlider, fmDepthLabel, fmDepthParameterID);
    sliderAndLabelSetup(treeState, fmFrequencySliderAttachment, fmFrequencySlider, fmFrequencyLabel, fmFrequencyParameterID);
    
    sliderAndLabelSetup(treeState, gainSliderAttachment, gainSlider, gainLabel, "GAIN");
    sliderAndLabelSetup(treeState, pitchSliderAttachment, pitchSlider, pitchLabel, "PITCH");

    pitchSlider.setTextValueSuffix(" Hz");
    fmFrequencySlider.setTextValueSuffix(" Hz");



}

OscillatorComponent::~OscillatorComponent()
{
}

void OscillatorComponent::paint(juce::Graphics& g)
{
    auto bounds = getLocalBounds().reduced(3);
    auto oscillatorLabelSpace = bounds.removeFromTop(18.0f);

   
    g.setColour(juce::Colours::white);
    g.setFont(18.0f);
    g.drawText("Main Oscillator", oscillatorLabelSpace, juce::Justification::centred);
    
    g.drawRoundedRectangle(bounds.getX(), bounds.getY(), bounds.getWidth(), (bounds.getHeight()/2) + 18.0f, 4.0f, 2.0f);
    g.drawRoundedRectangle(bounds.getX(), (bounds.getHeight() /2) + 18.0f, bounds.getWidth(), (bounds.getHeight() / 2) + 1.0f, 4.0f, 2.0f);

    auto fmLabelSpace = juce::Rectangle<float>(bounds.getX(), (bounds.getHeight() / 2) + 18.0f, bounds.getWidth(), 18.0f);
    g.drawText("Frequency Modulation Oscillator", fmLabelSpace, juce::Justification::centred);

    g.setFont(15.0f);
}

void OscillatorComponent::resized()
{
    //var to hold bounds of whole component, minus a bit
    const auto bounds = getLocalBounds().reduced(10);
    //var for distance between sliders
    const float sliderPadding = 10;
    //var for spacing slider evenly in box
    const float sliderWidth = (bounds.getWidth() / 2) - sliderPadding *2 ;
    const float sliderHeight = (bounds.getHeight() * 0.32); //just under half of window tall

    //x,y point for first fm slider
    const float sliderStartPointX = bounds.getX();
    const float fmsliderStartPointY = (fmOscillatorWaveSelector.getBottom() + 10);
    //y point for top row of sliders
    const float gainPitchSliderStartPointY = (mainOscillatorWaveSelector.getBottom() + 20);



    mainOscillatorWaveSelector.setBounds(bounds.getX() + bounds.getWidth() / 4, 25, bounds.getWidth() / 2, 20);
    //draw fm selector on bottom part of window
    fmOscillatorWaveSelector.setBounds(bounds.getX() + bounds.getWidth() /4, gainSlider.getBottom() + 36, bounds.getWidth() / 2, 20);//use pther slider bottoms

    //set bounds of leftmost sliders
    fmDepthSlider.setBounds(sliderStartPointX, fmsliderStartPointY, sliderWidth, sliderHeight);
    gainSlider.setBounds(sliderStartPointX, gainPitchSliderStartPointY, sliderWidth, sliderHeight);

    //set other sliders using previous's x position 
    fmFrequencySlider.setBounds(fmDepthSlider.getRight() + sliderPadding, fmsliderStartPointY, sliderWidth, sliderHeight);
    pitchSlider.setBounds(gainSlider.getRight() + sliderPadding, gainPitchSliderStartPointY, sliderWidth, sliderHeight);


    //draw labels relative to sliders (center bottom) and leave room betweeen them and comboboxes
    fmDepthLabel.setBounds(fmDepthSlider.getX(), fmOscillatorWaveSelector.getBottom() + 5, fmDepthSlider.getWidth(), 15);
    fmFrequencyLabel.setBounds(fmFrequencySlider.getX(), fmOscillatorWaveSelector.getBottom() + 5, fmFrequencySlider.getWidth(), 15);
    gainLabel.setBounds(gainSlider.getX(), mainOscillatorWaveSelector.getBottom() + 5, gainSlider.getWidth(), 15);
    pitchLabel.setBounds(pitchSlider.getX(), mainOscillatorWaveSelector.getBottom() + 5, pitchSlider.getWidth(), 15);

}


void OscillatorComponent::sliderAndLabelSetup(juce::AudioProcessorValueTreeState& treeState, std::unique_ptr<juce::AudioProcessorValueTreeState::SliderAttachment>& sliderAttachment,
    juce::Slider& slider, juce::Label& label, const juce::String parameterID) {
    //slider attachments

    //initialize with treestate etc
    sliderAttachment = std::make_unique<juce::AudioProcessorValueTreeState::SliderAttachment>(treeState, parameterID, slider);
    //===================================================================================================
    //slider

    //style sliders
    slider.setSliderStyle(juce::Slider::SliderStyle::RotaryHorizontalVerticalDrag);

    //set rotary parameters to change draw angle 
    ////rotary params radians measured from top of circle, instead of right like usual
    //// need larger end angle or else slider drawn in reverse, 
    float newStartAngleRadians = juce::MathConstants<float>::pi * 0.0f;
    float newEndAngleRadians = juce::MathConstants<float>::pi * 1.5f;

    slider.setRotaryParameters(newStartAngleRadians, newEndAngleRadians, true);

    //set text boxes
    slider.setTextBoxStyle(juce::Slider::TextBoxBelow, true, 50, 25);

    //add and make visible
    addAndMakeVisible(slider);
    //===================================================
    //label

    //set text
    label.setFont(17.0f);
    //attach to slider
    label.attachToComponent(&slider, false);
    //set justification type to make sure it is in middle
    label.setJustificationType(juce::Justification::centred);
    //make visible
    addAndMakeVisible(label);

}