
#include "PluginProcessor.h"
#include "PluginEditor.h"
#include "SynthVoice.h"

//==============================================================================
LastSynthesizerAudioProcessor::LastSynthesizerAudioProcessor()
#ifndef JucePlugin_PreferredChannelConfigurations
    : AudioProcessor(BusesProperties()
#if ! JucePlugin_IsMidiEffect
#if ! JucePlugin_IsSynth
        .withInput("Input", juce::AudioChannelSet::stereo(), true)
#endif
        .withOutput("Output", juce::AudioChannelSet::stereo(), true)
#endif
    ), myTreeState(*this, nullptr, "Parameters", createParameterLayout())//construct tree state
#endif
{

    //no garbage/leftover samples from last note (or last instance of synth)
    mySynth.clearVoices();
    mySynth.addVoice(new SynthVoice()); //


    mySynth.clearSounds();
    mySynth.addSound(new SynthSound());
}

LastSynthesizerAudioProcessor::~LastSynthesizerAudioProcessor()
{
    //juce synthesizer object handles voice and sound, deletes when no longer needed
}

//==============================================================================
const juce::String LastSynthesizerAudioProcessor::getName() const
{
    return JucePlugin_Name;
}

bool LastSynthesizerAudioProcessor::acceptsMidi() const
{
#if JucePlugin_WantsMidiInput
    return true;
#else
    return false;
#endif
}

bool LastSynthesizerAudioProcessor::producesMidi() const
{
#if JucePlugin_ProducesMidiOutput
    return true;
#else
    return false;
#endif
}

bool LastSynthesizerAudioProcessor::isMidiEffect() const
{
#if JucePlugin_IsMidiEffect
    return true;
#else
    return false;
#endif
}

double LastSynthesizerAudioProcessor::getTailLengthSeconds() const
{
    return 0.0;
}

int LastSynthesizerAudioProcessor::getNumPrograms()
{
    return 1;   //some hosts don't cope very well if you tell them there are 0 programs,
                // so this should be at least 1, even if you're not really implementing programs.
}

int LastSynthesizerAudioProcessor::getCurrentProgram()
{
    return 0;
}

void LastSynthesizerAudioProcessor::setCurrentProgram(int index)
{
}

const juce::String LastSynthesizerAudioProcessor::getProgramName(int index)
{
    return {};
}

void LastSynthesizerAudioProcessor::changeProgramName(int index, const juce::String& newName)
{
}

//==============================================================================
void LastSynthesizerAudioProcessor::prepareToPlay(double sampleRate, int samplesPerBlock)
{
    // Use this method as the place to do any pre-playback
    // initialisation that you need..
        //ignore unused samples from last key(s) pressed
    juce::ignoreUnused(samplesPerBlock);

    //so it is not constantly changing
    mySynth.setCurrentPlaybackSampleRate(sampleRate);

    for (int i = 0; i < mySynth.getNumVoices(); i++) {
        //getVoice returns juce SynthesizerVoice, I want our SynthVoice
        if (auto voice = dynamic_cast<SynthVoice*>(mySynth.getVoice(i)))
        {
            auto& gainValue = *myTreeState.getRawParameterValue("GAIN");
            auto& pitchValue = *myTreeState.getRawParameterValue("PITCH");
            voice->prepareToPlay(pitchValue, gainValue, sampleRate, samplesPerBlock, getTotalNumOutputChannels());
        }
    }
    //filter.preparetoplay
    myStateVariableFilter.prepareToPlay(getTotalNumOutputChannels(), sampleRate, samplesPerBlock);
    mySecondFilter.prepareToPlay(getTotalNumOutputChannels(), sampleRate, samplesPerBlock);
}

void LastSynthesizerAudioProcessor::releaseResources()
{
    // When playback stops, you can use this as an opportunity to free up any
    // spare memory, etc.
}

#ifndef JucePlugin_PreferredChannelConfigurations
bool LastSynthesizerAudioProcessor::isBusesLayoutSupported(const BusesLayout& layouts) const
{
#if JucePlugin_IsMidiEffect
    juce::ignoreUnused(layouts);
    return true;
#else
    // This is the place where you check if the layout is supported.
    // In this template code I only support mono or stereo.
    // Some plugin hosts, such as certain GarageBand versions, will only
    // load plugins that support stereo bus layouts.
    if (layouts.getMainOutputChannelSet() != juce::AudioChannelSet::mono()
        && layouts.getMainOutputChannelSet() != juce::AudioChannelSet::stereo())
        return false;

    // This checks if the input layout matches the output layout
#if ! JucePlugin_IsSynth
    if (layouts.getMainOutputChannelSet() != layouts.getMainInputChannelSet())
        return false;
#endif

    return true;
#endif
}
#endif

void LastSynthesizerAudioProcessor::processBlock(juce::AudioBuffer<float>& buffer, juce::MidiBuffer& midiMessages)
{
    juce::ScopedNoDenormals noDenormals;
    auto totalNumInputChannels = getTotalNumInputChannels();
    auto totalNumOutputChannels = getTotalNumOutputChannels();

    // In case I have more outputs than inputs, this code clears any output
    // channels that didn't contain input data, (because these aren't
    // guaranteed to be empty - they may contain garbage).

    // This is here to avoid people getting screaming feedback
    // when they first compile a plugin, but obviously you don't need to keep
    // this code if your algorithm always overwrites all the output channels.
    for (auto i = totalNumInputChannels; i < totalNumOutputChannels; ++i)
        buffer.clear(i, 0, buffer.getNumSamples());

    for (int i = 0; i < mySynth.getNumVoices(); i++) {
        //iterate through voices, check if synthvoice, and update their params.
            //can't call updateADSR() if they are not a synthvoice
        if (auto voice = dynamic_cast<SynthVoice*>(mySynth.getVoice(i))) {


            //getRawParamVal() return std::atomic<float>*, we need const floats so we dereference

            //get wave type from tree state
            const int& waveType = *myTreeState.getRawParameterValue("WAVETYPE");
            voice->getOscillator().setOscillatorWaveType(waveType);
            //this calls synthvoice::getOscilator() to get reference to my oscillator object. sets wave type to whatever is in tree state

            //get pitch value and set synth voices pitch
            const float& pitchValue = *myTreeState.getRawParameterValue("PITCH");
            voice->updatePitch(pitchValue);

            //get FM wave type and set synthvoice's oscillators' fmOscillator to it
            const int& fmWaveType = *myTreeState.getRawParameterValue("FMWAVETYPE");
            voice->getOscillator().setFMOscillatorWaveType(fmWaveType);

            //get fm params from tree state
            const float& fmDepth = *myTreeState.getRawParameterValue("MODDEPTH");
            const float& modulationFrequency = *myTreeState.getRawParameterValue("MODFREQ");
            //call oscilator object setfmparams()
            voice->getOscillator().setFrequencyModulationParameters(modulationFrequency, fmDepth);

            //get values of Amplitude Modulation parameters from valuetreestate
            const float& attack = *myTreeState.getRawParameterValue("ATTACK");
            const float& decay = *myTreeState.getRawParameterValue("DECAY");
            const float& sustain = *myTreeState.getRawParameterValue("SUSTAIN");
            const float& release = *myTreeState.getRawParameterValue("RELEASE");
            
            voice->updateADSR(attack, decay, sustain, release);
            //voice->updateADSR() calls synthVoice's updateADSR() which calls the AdsrData object's update().

            //get gain value and update synthvoice's gain object
            const float& gainValue = *myTreeState.getRawParameterValue("GAIN");
            voice->updateGain(gainValue);
        }
    }

    
    mySynth.renderNextBlock(buffer, midiMessages, 0, buffer.getNumSamples());
    /*
    creates the next block of audio output.
            **calls juce::Synthesiser::ProcessNextBlock() which calls RenderVoices(), which iterates through our SynthVoices and
            calls their version of RenderNextBlock().

            //midiIterator in processBlock() looks at where in buffer we were when key was hit, applies offset to account.
            //the rendervoices() call will have a startSample value of wherever new key was pressed, and a numSamples value
            //of (buffer length)-(index at which new key was pressed)---(256-startSample)

    FROM DOCS: This will process the next numSamples of data from all the voices, and add that output to the audio block supplied,
    starting from the offset specified. Note that the data will be added to the current contents of the buffer,
    so you should clear it before calling this method if necessary.

    The midi events in the inputMidi buffer are parsed for note and controller events, and these are used to trigger the voices.
     Note that the startSample offset applies both to the audio output buffer and the midi input buffer, so any midi events with
     timestamps outside the specified region will be ignored.

    */


    //filter sound of main output, not in each synth voice
    
    //get filter type from tree state
    const int& filterType = *myTreeState.getRawParameterValue("FILTERTYPE");

    //get values of parameters from valuetreestate
    const float& filterCutoff = *myTreeState.getRawParameterValue("FILTERCUTOFF");
    const float& filterResonance = *myTreeState.getRawParameterValue("FILTERRESONANCE");
    myStateVariableFilter.updateFilterParameters(filterType, filterResonance, filterCutoff);

    const int& filter2Type = *myTreeState.getRawParameterValue("FILTER2TYPE");

    //get values of parameters from valuetreestate
    const float& filter2Cutoff = *myTreeState.getRawParameterValue("FILTER2CUTOFF");
    const float& filter2Resonance = *myTreeState.getRawParameterValue("FILTER2RESONANCE");
    mySecondFilter.updateFilterParameters(filter2Type, filter2Resonance, filter2Cutoff);


    //process with filters after updating params
    myStateVariableFilter.process(buffer);
    mySecondFilter.process(buffer);
}


//==============================================================================
bool LastSynthesizerAudioProcessor::hasEditor() const
{
    return true;
}

juce::AudioProcessorEditor* LastSynthesizerAudioProcessor::createEditor()
{
    return new LastSynthesizerAudioProcessorEditor(*this);
}

//==============================================================================
void LastSynthesizerAudioProcessor::getStateInformation(juce::MemoryBlock& destData)
{

}

void LastSynthesizerAudioProcessor::setStateInformation(const void* data, int sizeInBytes)
{

}

//==============================================================================
// This creates new instances of the plugin..
juce::AudioProcessor* JUCE_CALLTYPE createPluginFilter()
{
    return new LastSynthesizerAudioProcessor();
}

juce::AudioProcessorValueTreeState::ParameterLayout LastSynthesizerAudioProcessor::createParameterLayout() {
    //make vector of parameters
    std::vector<std::unique_ptr<juce::RangedAudioParameter>> parameters;

    // all "tunable" parameter value ranges use optional 3rd arg for normailisableRange{} intervalValue to keep at certain amount of decimal places

    //add gain and pitch
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("GAIN", "Synthesizer Audio Gain", juce::NormalisableRange<float> {0.0f, 1.0f, 0.01f}, 0.2f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("PITCH", "Synthesizer Pitch Change", juce::NormalisableRange<float> {-300.0f, 300.0f, 1.0f}, 0.0f));


    //add oscilator select
    parameters.push_back(std::make_unique<juce::AudioParameterChoice>("WAVETYPE", "Oscillator Wave Type", juce::StringArray{ "Sine","Square","Saw" }, 0));

    

    //filter modifiers
        //statevariabletptfilter type = Enumerator-lowpass 0, bandpass 1, highpass 2
        //resonance - The bandwidth of the resonance increases with the value of the parameter. 
            //To have a standard 12 dB / octave filter, the value must be set at 1 / sqrt(2).
    parameters.push_back(std::make_unique<juce::AudioParameterChoice>("FILTERTYPE", "State Variable Filter Type", juce::StringArray{ "Low-Pass","Band-Pass","High-Pass" }, 0));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("FILTERRESONANCE", "Filter Resonance", juce::NormalisableRange<float> {0.1f, 2.0f, 0.1f }, 1.0f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("FILTERCUTOFF", "Filter Cutoff Frequency", juce::NormalisableRange<float> {20.0f, 20000.0f, 0.1f, 0.55f }, 300.0f));
    //cutoff min is 20 Hz, min of human hearing. skew factor of 0.6

    parameters.push_back(std::make_unique<juce::AudioParameterChoice>("FILTER2TYPE", "State Variable Filter 2 Type", juce::StringArray{ "Low-Pass","Band-Pass","High-Pass" }, 0));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("FILTER2RESONANCE", "Filter 2 Resonance", juce::NormalisableRange<float> {0.1f, 2.0f, 0.1f }, 1.0f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("FILTER2CUTOFF", "Filter 2 Cutoff Frequency", juce::NormalisableRange<float> {20.0f, 20000.0f, 0.1f, 0.55f }, 300.0f));



// add Amplitude Modulation params
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("ATTACK", "Attack", juce::NormalisableRange<float> {0.1f, 1.0f, 0.01f }, 0.1f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("DECAY", "Decay", juce::NormalisableRange<float> {0.1f, 2.0f, 0.01f }, 0.1f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("SUSTAIN", "Sustain", juce::NormalisableRange<float> {0.1f, 1.0f, 0.01f }, 1.0f));
    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("RELEASE", "Release", juce::NormalisableRange<float> {0.1f, 3.0f, 0.01}, 0.6f));

    // freq modulation params
    parameters.push_back(std::make_unique<juce::AudioParameterChoice>("FMWAVETYPE", "FM Wave Type", juce::StringArray{ "Sine","Square","Saw" }, 0));

    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("MODFREQ", "FM Frequency", juce::NormalisableRange<float> {0.f, 900.0f, 0.01f, 0.2f }, 0.0f));
    // first param = min frequency---> 0.0 freq on modulation oscilator would result in no change
        //second param is max freq e.g. 1000Hz

    parameters.push_back(std::make_unique < juce::AudioParameterFloat>("MODDEPTH", "FM Depth", juce::NormalisableRange<float> {0.f, 700.0f, 0.01f, 0.6f }, 0.0f));
    //depth is sort of like gain on modulation wave. How big(tall) I want fm wave to be. (normally in sine (-1)-(1) is depth of 1)
        //an FM depth of 1 is a tiny change. goes up or down 1 Hz. ***make sure result is not negative on lower end. (original - depth)

    return{ parameters.begin(), parameters.end() };
}
