
#include "SynthVoice.h"
#include <JuceHeader.h>
#include <assert.h>

//Must return true if this voice object is capable of playing the given sound (sound is properly loaded).
bool SynthVoice::canPlaySound(juce::SynthesiserSound* sound) {
        //try to cast sound as SynthSound, if cast is sucessful, true. If cast fails, false, cant play sound
    return dynamic_cast<juce::SynthesiserSound*>(sound) != nullptr;

}

void SynthVoice::startNote(int midiNoteNumber, float velocity, juce::SynthesiserSound* sound, int currentPitchWheelPosition) {
    //Called to start a new note.

    //set osc base freq
    myOscillatorObject.setWaveFrequency(midiNoteNumber);
    //reset frequency using pitch value;
    auto newFrequency = myOscillatorObject.getFrequency() + voiceNewPitchValue;


    //use native adsr object method through data object
    myAdsrObject.noteOn();
}


void SynthVoice::stopNote(float velocity, bool allowTailOff) {
    //Called to stop a note.

    //use native adsr object method through data object
    myAdsrObject.noteOff();

    //be double sure note gets cleared
    if (myAdsrObject.isActive() == false || allowTailOff == false) clearCurrentNote();
}


void SynthVoice::pitchWheelMoved(int newPitchWheelValue) {}//part of juce object, not used

void SynthVoice::controllerMoved(int controllerNumber, int newControllerValue) {}//part of juce object, not used



void SynthVoice::prepareToPlay(float pitchValue, float gainValue, double sampleRate, int samplesPerBlock, int outputChannels) { 
    // Use this method as the place to do any pre-playback
  // initialisation that you need..
    
    //set ADSR object's samplerate
    myAdsrObject.setSampleRate(sampleRate);


  //create processspec to neatly pass info into oscilator
    juce::dsp::ProcessSpec mySpec;
    mySpec.maximumBlockSize = samplesPerBlock;
    mySpec.sampleRate = sampleRate;
    mySpec.numChannels = outputChannels;

    myOscillatorObject.prepareToPlay(mySpec);

    //prepare gain
    myGain.prepare(mySpec);

    //set base gain with linear val, not converted to dB (0-1)
    updateGain(gainValue);

    updatePitch(pitchValue);


    //let renderNextBlock know it is prepared
    hasBeenPrepared = true;
}

void SynthVoice::updateADSR(const float attackValue, const float decayValue, const float sustainValue, const float releaseValue) {
    //call update in my data object
    myAdsrObject.update(attackValue, decayValue, sustainValue, releaseValue);

}
void SynthVoice::updateGain(const float gainValue) {
    myGain.setGainLinear(gainValue);
}
void SynthVoice::updatePitch(const float pitchValue) {
    voiceNewPitchValue = pitchValue;
}


void SynthVoice::renderNextBlock(juce::AudioBuffer< float >& outputBuffer, int startSample, int numSamples) {


    //jassert prepared. if false stop, because block hasnt been prepared. 
    jassert(hasBeenPrepared = true);

    //If the voice is currently silent, it should just return without doing anything. means SynthSound has not been loaded
    if (isVoiceActive() == false) return;

    clickFixBuffer.setSize(outputBuffer.getNumChannels(), numSamples, false, false, true);

    //*********
    // This can grow/shrink the buffer's length and add/remove channels.
    //Normally, this type of thing wouldn't be allowed in something that's being called by the audio thread.
    //Juce helps us, they made the audio buffer not allocate every time I go through the buffer, unless it needs
    //to allocate a bit more memory for the size of buffer needed
            //setsize has 3 optional flags, one of which is bool avoidReallocating. This avoids shrinking memory allocated, 
            //but will still increase memory if new buffer is bigger than the current one
         //bool keepExistingContent will try to preserve as much of the old data as it can in the new buffer (I dont want because I am replacing)
        //bool 	clearExtraSpace = false clears any extra channels or space that is allocated. If false, then this space is left uninitialised.

    //clear the new buffer each call
    clickFixBuffer.clear();



   
//for any juce dsp module methods, I need to pass it an audio block (I am using my new one instead of outputbuffer now)
    juce::dsp::AudioBlock<float> audioBlock{ clickFixBuffer };
//Creates an AudioBlock that points to the data in an AudioBuffer.

    //AudioBlock does not copy nor own the memory pointed to by dataToUse.Therefore it's my
    //responsibility to ensure that the buffer is retained throughout the life-time of the AudioBlock without being modified.

    //process my block through my oscilator
    //Processes the input and output buffers supplied in the processing context
    
    myOscillatorObject.setFrequency(myOscillatorObject.getFrequency() + voiceNewPitchValue);
    myOscillatorObject.getNextAudioBlock(audioBlock);

    //process with gain (on newly changed audio block)
    myGain.process(juce::dsp::ProcessContextReplacing<float>(audioBlock));
    //audioBlock is just an alias for the outputBuffer, so outputBuffer will contain all the audio data
    // OBJ process context replacing = when I pass buffer/audio block into process method, the result will REPLACE
    //whatever was in the buffer before. 
    
    // 
    //apply envelope to buffer (as opposed to going sample by sample) used to be on output buffer
    myAdsrObject.applyEnvelopeToBuffer(clickFixBuffer, 0, clickFixBuffer.getNumSamples());
    //0 because I always want beginning of clickFixBuffer


    //below added for click fix
    for (int channelNumber = 0; channelNumber < outputBuffer.getNumChannels(); channelNumber++) {
        outputBuffer.addFrom(channelNumber, startSample, clickFixBuffer, channelNumber, 0, numSamples);

        //addfrom() takes care of necessary readjusting to avoid clicks in transitions
        //addFrom(int destChannel, int destStartSample, const AudioBuffer & source, int sourceChannel, int sourceStartSample, int numSamples)

        //because of how the processor does synth.rendernextblock, i figured start sample would be 0 and numsamples would be whatever is passed
        //not always the case.
    }

    //If the sound that the voice is playing finishes during the course of this rendered block, it must call clearCurrentNote(), to tell 
    //the synthesiser that it has finished.
    if (myAdsrObject.isActive() == false) clearCurrentNote();

}


//********************************************************************************************************************************
/*clicking:
thought that when the processor called rendernext(), the whole audio buffer would be processed from sample 0 to end (getnumsamples())
    set up synth voice render method incorrectly.
    both key press and release send midi messages. but this can happen at any point in buffer, not just first sample
    needs to know where in buffer key has been pressed/released etc at appropriate position in buffer, not just
    from 0 to numSamples();
    offsetting happens in buffer when I get midi messages in

**output buffer into renderNextBlock, something may already be happening in process block to audio.
 -if it's in release stage, and I need to play another note, if I apply ADSR envelope to same buffer,
 the phase of my sine wave isnt at 0, it would be negative. disjunction in phase time to next note.
 Starting a note while the last one has note hasnt finished. the phase numbers have a discrepancy
 ->no smooth transition between 2 sine notes
 SOLVE:
 create another audio buffer and apply adsr,gain,oscilator stuff. add to main output buffer.
    if output buffer still has audio in it, it will change positions to avoid the click

    SynthVoice render() "the output audio data must be added to the current contents of the buffer provided."

*/

